`timescale 10ns / 10ps

`define VSYNCSTART			0
`define VSYNCEND				2
`define HSYNCSTART			0
`define HSYNCEND				96
`define HTOTAL					800
`define HBACKPORCH			64
`define VTOTAL50				588
`define VTOTAL60				525
`define HACTSTART				144
`define HACTEND				784
`define VACTSTART				35
`define VACTEND				515
`define VBACKPORCH			48
`define VACTPAPER				467
`define FLASHSPEED			19
`define ZX_RAM_START			16384
`define ZX_MAIN_MEM			23296 
`define VIDEO_PIX_INT		0
`define RAM_PAGE0				16384
`define RAM_PAGE1				32768
`define RAM_PAGE2				49152


module ZX_Spectrum128
(
	input           	clk50,           			// system clock 50Mhz on board
	output reg			led0,							// Led 0, debug
	input 				vsync_select,				// SW0 to select 50 or 60Hz
	input 				turbo_mode,					// 3.5-7Mhz switch
	input           	rst_n,         			// reset, low active
	input					rst_z80,						// Reset for Z80
	input	 	 			keyb_in,						// keyb_in, keyboard in
	input		 			tapein, 						// ZX Spectrum tape in
	input     [4:0]	KempstonJoystick,			// Kempston Joystick
	output reg[47:0]	tstDATA_LED,				// 7 seg LED data
	output  				vga_hsync,					// Out h sync
	output  				vga_vsync,    				// Out v sync
	output reg 			sound,						// ZX Spectrum sound
	output reg[3:0]	out_R,						// Output RED
	output reg[3:0]	out_G,						// Output GREEN
	output reg[3:0]	out_B,  						// Output BLUE
	output 				audio_out					// Audio output
);

reg bPrevTin;											// Previous tape in state
reg bTinLock;											// Tape in locked
reg [21:0] rCounterUnlockTape;					// Timer

reg clk25;												// 25M clock
reg [23:0] BlLedDivider;							// Blinking LED, for fun
reg [5:0]  DigSelect;								//	Digit selector for 6x(7+1) LED indicator
reg [7:0]  CurDigit;									//	Current digit bit by bit
reg [3:0]  NextDigit;								//	Current digit to display in HEX	
reg [23:0] DataToDisplay;							//	Entire 2 bit number to display

/* Z80 pins */
wire 		a_nM1;										// It indicates the Machine Cycle number 1. So the CPU activates this output, when 
															// the first machine cycle of an instruction is in progress
															
wire 		a_nMREQ;										// The memory request pin used to return low (0) signal, when communication with memory blocks is desired.
wire 		a_nIORQ;										// The Input Output request pin is used to return low (0) signal when communication with IO ports is desired.
wire 		a_nRD;										// Read signal
wire 		a_nWR;										// Write signal
wire 		a_nRFSH;										// The RFSH and MREQ both can be used to indicate the lower seven bits of the address bus can be used as 
															// refresh address for Dynamic memories.
															
wire 		a_nHALT;										// The Halt state indicates the CPU has executed a HALT instruction and waiting for interrupt.
wire 		a_nBUSACK;									// It is Bus Acknowledge pin. It indicates the requesting device that the address bus, data bus, 
															// control bus enter into a high impedance state.

reg 		a_nWAIT;										// This signal is used for the communication between CPU and memory units or IO devices. Until the signal 
															// is inactive the CPU continuously enters the Wait state.
															
reg 		a_nINT;										// Interrupt Request pin is used to generate interrupts by IO devices. After executing each instruction, CPU 
															// checks for the Interrupt requests
reg 		a_nNMI;										// The Non-Maskable Interrupt pin is negative edge triggered. Its priority is higher than the INT. It does 
															// not depend on the status of the interrupt enable flip-flops. It automatically forces the CPU to 
															// restart at location 0066H.

reg 		a_nBUSRQ;									// Bus request Pin forces the address bus, data bus and control bus to enter a high impedance state. 
															// It is always recognized at the end of each machine cycle.

wire [7:0] a_D;										// Data bus for Z80 CPU
wire[15:0] wZXGlobalAddress;						// Global Z80 memory address
reg [7:0]  wZXGlobalDataO;							// Global memory data out

reg [14:0] aROM_addr;								// ZX ROM address, 15 bits, two pages by 16384 bytes (required for 128K version)
wire [7:0] aROM_q;									// ROM data

reg [7:0] r8_128KSwitch;							// Port 7FFD to manage 128K

reg [13:0] aVRAM_addr_a0;							// RAM bank 0, adress lines A
reg [13:0] aVRAM_addr_b0;							// address lines B
reg [7:0]  aVRAM_data_a0;							// Data in A
reg [7:0]  aVRAM_data_b0;							// Data in B

wire [7:0] aVRAM_q_a0;								// Data out A
wire [7:0] aVRAM_q_b0;								// Data out B
reg        aVRAM_wren_a0;							// WREN A
reg        aVRAM_wren_b0;							// WREN B

reg [13:0] aVRAM_addr_a1;							// Data bank 1, address lines A
reg [13:0] aVRAM_addr_b1;							// Same as above
reg [7:0]  aVRAM_data_a1;							// Same as above
reg [7:0]  aVRAM_data_b1;							// Same as above

wire [7:0] aVRAM_q_a1;								// Same as above
wire [7:0] aVRAM_q_b1;								// Same as above
reg        aVRAM_wren_a1;							// Same as above
reg        aVRAM_wren_b1;							// Same as above

reg [13:0] aVRAM_addr_a2;							// Data bank 2, address lines A
reg [13:0] aVRAM_addr_b2;							// Same as above
reg [7:0]  aVRAM_data_a2;							// Same as above
reg [7:0]  aVRAM_data_b2;							// Same as above

wire [7:0] aVRAM_q_a2;								// Same as above
wire [7:0] aVRAM_q_b2;								// Same as above
reg        aVRAM_wren_a2;							// Same as above
reg        aVRAM_wren_b2;							// Same as above

reg [13:0] aVRAM_addr_a3;							// Same as above
reg [13:0] aVRAM_addr_b3;							// Same as above
reg [7:0]  aVRAM_data_a3;							// Same as above
reg [7:0]  aVRAM_data_b3;							// Same as above

wire [7:0] aVRAM_q_a3;								// Same as above
wire [7:0] aVRAM_q_b3;								// Same as above
reg        aVRAM_wren_a3;							// Same as above
reg        aVRAM_wren_b3;							// Same as above

reg [13:0] aVRAM_addr_a4;							// Same as above
reg [13:0] aVRAM_addr_b4;							// Same as above
reg [7:0]  aVRAM_data_a4;							// Same as above
reg [7:0]  aVRAM_data_b4;							// Same as above

wire [7:0] aVRAM_q_a4;								// Same as above
wire [7:0] aVRAM_q_b4;								// Same as above
reg        aVRAM_wren_a4;							// Same as above
reg        aVRAM_wren_b4;							// Same as above

reg [13:0] aVRAM_addr_a5;							// Same as above
reg [13:0] aVRAM_addr_b5;							// Same as above
reg [7:0]  aVRAM_data_a5;							// Same as above
reg [7:0]  aVRAM_data_b5;							// Same as above

wire [7:0] aVRAM_q_a5;								// Same as above
wire [7:0] aVRAM_q_b5;								// Same as above
reg        aVRAM_wren_a5;							// Same as above
reg        aVRAM_wren_b5;							// Same as above

reg [13:0] aVRAM_addr_a6;							// Same as above
reg [13:0] aVRAM_addr_b6;							// Same as above
reg [7:0]  aVRAM_data_a6;							// Same as above
reg [7:0]  aVRAM_data_b6;							// Same as above

wire [7:0] aVRAM_q_a6;								// Same as above
wire [7:0] aVRAM_q_b6;								// Same as above
reg        aVRAM_wren_a6;							// Same as above
reg        aVRAM_wren_b6;							// Same as above

reg [13:0] aVRAM_addr_a7;							// Same as above
reg [13:0] aVRAM_addr_b7;							// Same as above
reg [7:0]  aVRAM_data_a7;							// Same as above
reg [7:0]  aVRAM_data_b7;							// Same as above

wire [7:0] aVRAM_q_a7;								// Same as above
wire [7:0] aVRAM_q_b7;								// Same as above
reg        aVRAM_wren_a7;							// Same as above
reg        aVRAM_wren_b7;							// Same as above

reg 		  PreviousVsync;							// Previous Vsync state, required to produce an exact timing
reg [16:0] sp50Hzcntr;								// Counter of clocks to generate an interrupt

reg bDrawingBorder;									// Flag that tells us that currently we are drawing border

wire 		  VideoClock;								// Actual pixelclock which will be selected between 46 and 50MHz to have 50 or 60 fps
															// Pixelclock is selected based on vsync_select
assign VideoClock = vsync_select? ClkVid:clk50;
wire 		  ClkVid;									// Video pixelclock 46MHz
reg		  outVsync;									// Output vsync
reg 		  outHsync;									// Output hsync
assign vga_hsync = outHsync;						// Assign output hsync pin to signal name
assign vga_vsync = outVsync;        			// Assign output vsync pin to signal name
reg 		  rZXFlashAttr;							// Flash-attribute
reg [4:0]  rZXFlashCntr;							// Counter for Flash attribute
reg		  outVideoPCLK;							// VGA pclk counter
reg		  [9:0]outPixCntr;						// Pixels in line counter
reg		  [9:0]outLineCntr;						// Lines in frame counter
reg [23:0] pixeladdr;								// Pixel address before transferring into ZX format
reg [6:0]  ColorLineAddr;							// Color attribute counter (1 line only), range 0..31, used in calculating VRAM memory address
reg [255:0]ColorDataMemCpy;						// Fetched data from color memory for 1 line only
reg [7:0]  CurColorAttribute;						// Current color attribute
reg [6:0]  CurColorLineAddr;						// Address for an attrubute cache
reg 		  ColorSync;								// Signal to make proper timing to fetch data from FPGA RAM
reg [2:0] rBorderColor;								// Border color

reg [7:0] r8Uart115RxState;                  // RX state machine                             
reg [7:0] r8Uart115RxNextState;              // RX next state                                
reg [7:0] r8Uart115RxDelay;                  // RX delay                                     
reg [3:0] r4Bits115RxCount;                  // RX bits counter                              
reg [7:0] r8Uart115RxByte;                   // RX byte                                          
reg		 bRx115ShiftReq;                    // Bit shift request                            
reg [7:0] rUdata115Out;                      // Received byte                                
reg bkeyb_in;											// Register to hold input value for keyboard input to prevent internal overload
reg [7:0] rx_byte_cntr;								// Byte counter of external stream
reg [39:0]zx_Keyboard_state;						// State of ZX keyboard
reg [7:0] r8Key1;										// First pressed key
reg [7:0] r8Key2;										// Second pressed key
reg [7:0] r8Key3;										// Third pressed key

wire rst_g;												// Reset is either z80 or global. Do ne as a sligh mod to get code more in style
assign rst_g = rst_n & rst_z80;					// If either reset or reset z80 eq 0 then global reset is 0

reg [7:0] AY8910_d_in;								// AY 8910 data input
wire[7:0] AY8910_d_out;								// AY 8910 data output
wire		 AY8910_d_oe_l; 							// Output signal (not used)
wire 		 AY8910_BDIR;								// BDIR input
wire 		 AY8910_BC1;								// BC 1 input
wire [7:0]wSound8b;									// 8 bit sound
wire 		 audio_clock;								// 28M audio clock
wire		 c80M;										// 80M clock for audio pwm

reg [2:0] clk_psg_r;									// Sync for AY8910
reg en_clk_psg_i;										// Sync for AY8910
wire [9:0] audio_out9;								// 10 bit output of AY8910
assign wSound8b[7:0] = audio_out9[9:2];		// 10 bit to 8 bit DAC mapping

/* 

The following logic has been implemented for AY8910:

BDIR = not (a1 or nWR) provided that nIRQ = 0 
BDIR = 0 if nIRQ = 1

BC1 = not (a1 or not a14) provided that nIRQ = 0
BC1 = 0 if nIRQ = 1

*/

assign AY8910_BDIR  = a_nIORQ ? 0 : !(wZXGlobalAddress[1] | a_nWR);
assign AY8910_BC1 = a_nIORQ ? 0 : !(wZXGlobalAddress[1] | !wZXGlobalAddress[14]);

															// Clock to be generated
															// 3p5M 		--CPU
															// 7M			--RAM controller
															// 28M		--Audio 8912	
															// 46.296296--Video	(25/27 in pll settings)
															// 80M		--Audio PWM out


wire		rMainClkIn3p5;								// 3p5 clock for the cpu
wire		rMainClkIn28p0;							// 28p0 clock for cpu
wire     MainClkMem56p0;							// 56p0 for the memory

wire 		rCpuClkMain;								// 3.5MHz clock for CPU
wire     mc_clk;                             // Memory clock
reg 		aud_half_clock;							// 1.75MHz clock for audio	

															// Fast CPU clock if selector is set to fast mode
assign rCpuClkMain = rClkSwitch? rMainClkIn3p5:rMainClkIn28p0;
assign mc_clk = MainClkMem56p0;					// Memory clock always set to 56MHz

reg rClkSwitch;									   // Clock select signal
reg rClkSwitchPrev;									// Previous state of clock select
reg [23:0] rSwDelay;									// Delay to minimize glitches during clock switch

reg 		videoreset;									// Reset for videoadapter
reg 		prev60_50;									// Previous state of 50_60 switch

pll_main main_pll(
	.inclk0(clk50),									// Input is 50MHz
	.c0(rMainClkIn3p5),								// Output is 3.5MHz for Z80 clock
	.c1(rMainClkIn28p0),								// 28MHz for CPU
	.c2(MainClkMem56p0)								// 56MHz for memory controller
);


pll_video pll_vid(
	.inclk0(clk50),									// Input is 50MHz
	.c0(ClkVid) 										// Ouput is 46.296296MHz 
);

pll_aud80 pll_audio(
	.inclk0(clk50),									// Input is 50MHz
	.c0(c80M),											// 80MHz for PWM out
	.c1(audio_clock)									// Output is Audio clock (28M)
);										
															
clocks AudioClkGen(									// Audio clocks generator
	.CLK(audio_clock),								//:	28M in
	.nRESET(rst_g),									//:	1.75 out
	.CLKEN_PSG(aud_half_clock),
	.CLKEN_CPU(dummy_sig) 	//:	
	);


/*YM2149 AY8910(
  .I_DA(AY8910_d_in),                			// Audio data in
  .O_DA(AY8910_d_out),       		         	// Audio data out
  .I_A9_L(0),              						// A9_L
  .I_A8(1),                						// A8
  .I_BDIR(AY8910_BDIR),              			// BDIR
  .I_BC2(1),               						// BC2
  .I_BC1(AY8910_BC1),               			// BC1
  .I_SEL_L(0),             						// SEL_L

  .O_AUDIO(wSound8b),             				// Ouput
  .ENA(en_clk_psg_i),                 			// CE signal
  .RESET_L(rst_g),             					// reset
  .CLK(audio_clock)                 			// 28MHz
  );
*/

jt49_bus AY8910( 										// note that input ports are not multiplexed
    .rst_n(rst_g),									// reset
    .clk(audio_clock),    							// 28MHz signal on positive edge
    .clk_en(en_clk_psg_i),							// CE signal
								
    .bdir(AY8910_BDIR),								// BDIR
    .bc1(AY8910_BC1),								// BC1
    .din(AY8910_d_in),								// Databus in

	 .sel(0), 											// if sel is low, the clock is divided by 2
    .dout(AY8910_d_out),							// Data bus out
    .sound(audio_out9),  							// combined channel output
    .A(null1),      									// linearised channel output, A
    .B(null2),											// linearised channel output, B
    .C(null3),											// linearised channel output, C
    .sample(null4),									// Sample signal	
    .IOA_in(null5),									// IOA bus in
    .IOA_out(null6),									// IOA bus out
    .IOB_in(null7),									// IOB bus in
    .IOB_out(null8)									// IOB bus out
);
  
always @ (posedge audio_clock)               // CE signal generator for AY8910
begin
	if (clk_psg_r >= 5)								// Main clock has to be divided by 6
	begin
		clk_psg_r <= 0;								// Cycle through
	end else
	begin
		clk_psg_r <= clk_psg_r + 1;				// Counter to 6
	end
	
	if (clk_psg_r == 2)								// If counter is equal 2 then set CE to 1
	begin
		en_clk_psg_i <= 1;							// Set it to 1
	end else
	begin
		en_clk_psg_i <= 0;							// Otherwise set it to 0
	end
end 

  
/*
ym2149_audio AY8910(
   .clk_i(audio_clock),              //: in     std_logic -- system clock
   .en_clk_psg_i(en_clk_psg_i),       //: in     std_logic -- PSG clock enable
   .sel_n_i(0),            //: in     std_logic -- divide select, 0=clock-enable/2
   .reset_n_i(rst_g),          //: in     std_logic -- active low
   .bc_i(AY8910_BC1),               //: in     std_logic -- bus control
   .bdir_i(AY8910_BDIR),             //: in     std_logic -- bus direction
   .data_i(AY8910_d_in),             //: in     std_logic_vector(7 downto 0)
   .data_r_o(AY8910_d_out),           //: out    std_logic_vector(7 downto 0) -- registered output data
   .ch_a_o(null_sig1),             //: out    unsigned(11 downto 0)
   .ch_b_o(null_sig2),             //: out    unsigned(11 downto 0)
   .ch_c_o(null_sig3),             //: out    unsigned(11 downto 0)
   .mix_audio_o(audio_out14),        //: out    unsigned(13 downto 0)
   .pcm14s_o(null_sig4)           //: out    unsigned(13 downto 0)
);
*/

alx_audio alx_out(
	.bSys80MClock(c80M),								// System clock 
	.V8_DigInput(wSound8b),							// 8 bit input
	.rst_n(rst_g),										// reset
	.oPWM_P(audio_out)								// PWM out	
);
  
T80a aZ80(
		.RESET_n(rst_z80),							// Reset 
		.CLK_n(rCpuClkMain),							// Clock input
		.WAIT_n(a_nWAIT),								// WAIT input
		.INT_n(a_nINT),								// INT input
		.NMI_n(a_nNMI),								// NMI input
		.BUSRQ_n(a_nBUSRQ),							// BUSRQ INPUT
		.M1_n(a_nM1),									// M1 output
		.MREQ_n(a_nMREQ),								// MREQ output
		.IORQ_n(a_nIORQ),								// IORQ output
		.RD_n(a_nRD),									// RD output
		.WR_n(a_nWR),									// WR output
		.RFSH_n(a_nRFSH),								// RFSH output
		.HALT_n(a_nHALT),								// HALT output
		.BUSAK_n(a_nBUSACK),							// BUSACK output
		.A(wZXGlobalAddress),						// Address outout
		.D(a_D)											// Data inout
	);
	
/*
z80_top_direct_n aZ80(
    .nM1(a_nM1),										// M1 output
    .nMREQ(a_nMREQ),									// MREQ output
    .nIORQ(a_nIORQ),									// IORQ output
    .nRD(a_nRD),										// RD output
    .nWR(a_nWR),										// wr output	
    .nRFSH(a_nRFSH),									// RFSH output
    .nHALT(a_nHALT),									// HALT output
    .nBUSACK(a_nBUSACK),							// BUSACK output

    .nWAIT(a_nWAIT),									// WAIT input
    .nINT(a_nINT),									// INT input
    .nNMI(a_nNMI),									// NMI input
    .nRESET(rst_z80),								// Reset 
    .nBUSRQ(a_nBUSRQ),								// BUSRQ input

    .CLK(rCpuClkMain),								// Clock input
    .A(wZXGlobalAddress),							// Address outout
    .D(a_D)												// Data inout
);*/
															// Assignment of dara bus based on a_nWR signal		
assign a_D = a_nRD ? 8'bZZZZZZZZ:wZXGlobalDataO;

RAM16K aMEM0(											// Memory block 0
.address_a(aVRAM_addr_a0), 						// Address lines A	
.address_b(aVRAM_addr_b0),							// Address lines B
.clock(clk50),										   // Clock
.data_a(aVRAM_data_a0),							 	// Data in A
.data_b(aVRAM_data_b0),								// Data in B
.wren_a(aVRAM_wren_a0),								// WREN A
.wren_b(aVRAM_wren_b0),								// WREN B
.q_a(aVRAM_q_a0),										// Data out A
.q_b(aVRAM_q_b0)										// Data out B
);

RAM16K aMEM1(											// Same as above
.address_a(aVRAM_addr_a1), 						// Same as above	
.address_b(aVRAM_addr_b1),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a1),							 	// Same as above
.data_b(aVRAM_data_b1),								// Same as above
.wren_a(aVRAM_wren_a1),								// Same as above
.wren_b(aVRAM_wren_b1),								// Same as above
.q_a(aVRAM_q_a1),										// Same as above
.q_b(aVRAM_q_b1)										// Same as above
);

RAM16K aMEM2(											// Same as above
.address_a(aVRAM_addr_a2), 						// Same as above	
.address_b(aVRAM_addr_b2),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a2),							 	// Same as above
.data_b(aVRAM_data_b2),								// Same as above
.wren_a(aVRAM_wren_a2),								// Same as above
.wren_b(aVRAM_wren_b2),								// Same as above
.q_a(aVRAM_q_a2),										// Same as above
.q_b(aVRAM_q_b2)										// Same as above
);

RAM16K aMEM3(											// Same as above
.address_a(aVRAM_addr_a3), 						// Same as above	
.address_b(aVRAM_addr_b3),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a3),							 	// Same as above
.data_b(aVRAM_data_b3),								// Same as above
.wren_a(aVRAM_wren_a3),								// Same as above
.wren_b(aVRAM_wren_b3),								// Same as above
.q_a(aVRAM_q_a3),										// Same as above
.q_b(aVRAM_q_b3)										// Same as above
);

RAM16K aMEM4(											// Same as above
.address_a(aVRAM_addr_a4), 						// Same as above	
.address_b(aVRAM_addr_b4),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a4),							 	// Same as above
.data_b(aVRAM_data_b4),								// Same as above
.wren_a(aVRAM_wren_a4),								// Same as above
.wren_b(aVRAM_wren_b4),								// Same as above
.q_a(aVRAM_q_a4),										// Same as above
.q_b(aVRAM_q_b4)										// Same as above
);

RAM16K aMEM5(											// Same as above
.address_a(aVRAM_addr_a5), 						// Same as above	
.address_b(aVRAM_addr_b5),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a5),							 	// Same as above
.data_b(aVRAM_data_b5),								// Same as above
.wren_a(aVRAM_wren_a5),								// Same as above
.wren_b(aVRAM_wren_b5),								// Same as above
.q_a(aVRAM_q_a5),										// Same as above
.q_b(aVRAM_q_b5)										// Same as above
);

RAM16K aMEM6(											// Same as above
.address_a(aVRAM_addr_a6), 						// Same as above	
.address_b(aVRAM_addr_b6),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a6),							 	// Same as above
.data_b(aVRAM_data_b6),								// Same as above
.wren_a(aVRAM_wren_a6),								// Same as above
.wren_b(aVRAM_wren_b6),								// Same as above
.q_a(aVRAM_q_a6),										// Same as above
.q_b(aVRAM_q_b6)										// Same as above
);

RAM16K aMEM7(											// Same as above
.address_a(aVRAM_addr_a7), 						// Same as above	
.address_b(aVRAM_addr_b7),							// Same as above
.clock(clk50),										   // Same as above
.data_a(aVRAM_data_a7),							 	// Same as above
.data_b(aVRAM_data_b7),								// Same as above
.wren_a(aVRAM_wren_a7),								// Same as above
.wren_b(aVRAM_wren_b7),								// Same as above
.q_a(aVRAM_q_a7),										// Same as above
.q_b(aVRAM_q_b7)										// Same as above
);

ROMV1 aROM (											// ZX ROM address, 128K version
	.address(aROM_addr),								// Address bus
	.clock(clk50),										// Clock bus
	.q(aROM_q)											// Data out
);

reg [31:0]	ClkSwDebounce;

															// 50Hz-60Hz switch polling and if detected then reset the video adapter
always @ (posedge rMainClkIn3p5 or negedge rst_n)	
begin
	if (rst_n == 0)									// If reset
	begin
		videoreset <= 0;								// Reset the video 
		prev60_50 <= vsync_select;					// Copy current vsync select switch state
		ClkSwDebounce <= 0;
	end else
	begin
		if (prev60_50 != vsync_select)			// If state changed
		begin
			if (ClkSwDebounce > 7000000)			// 2S debounce (3.5M clk * 2)
			begin
				prev60_50 <= vsync_select;			// Remember the current state
				ClkSwDebounce <= 0;					// If over then reset debouncer
			end else
			begin
															// Increment debounce
				ClkSwDebounce <= ClkSwDebounce + 1;
			end
			videoreset <= 0;							// Activate reset for video
		end else
		begin
			videoreset <= 1;							// Otherwise de-activate reset for video
			ClkSwDebounce <= 0;						// And reset debounce for safety
		end 
	end
end
															// Polling clock switch
always @ (posedge rCpuClkMain or negedge rst_n)	
begin
	if (rst_n == 0)									// Reset is active?
	begin
		rClkSwitch <= 0;								// If yes, then reset all of the signals
		rSwDelay <= 0;									// Reset debounce
		rClkSwitchPrev <= 0;							// Reset previous state of selector
	end else
	begin
		if (rSwDelay == 16'hFFFF)					// If delay is expired (debounce)
		begin
															// If no RAM/ROM access and switch position changed
			if (a_nRD == 1 && a_nWR == 1 && rClkSwitchPrev != turbo_mode)
			begin
				rClkSwitchPrev <= turbo_mode;    // Set prvious state of turbo selector
				rSwDelay <= 0;							// Set debounce timer
				if (turbo_mode == 0)					// If turbo switch is off
				begin
					rClkSwitch <= 0;					// Then setting switch variable to 0
				end else
				begin
					rClkSwitch <= 1; 					// Otherwise setting switch variable to 1
				end
			end
		end else
		begin
			rSwDelay <= rSwDelay + 1;				// Increment debounce delay
		end
	end
end

always @ (posedge clk25 or negedge rst_n)		// Poll in the data for USB keyboard in order to go from signal to register
begin
	if (rst_n == 1'b0)   	                  // If reset, then initialize everything         
	begin
		bkeyb_in <= 0;									// Startup value
	end else
	begin
		bkeyb_in <= keyb_in;							// Polling
	end
end

always @(negedge clk25 or negedge rst_n)		//Uart RX procedure for keyboard @115200
begin
	if (rst_n == 1'b0)   	                  // If reset, then initialize everything         
	begin
		r8Uart115RxState <= 0;                 // RX state machine                             
		r8Uart115RxNextState <= 0;             // RX next state                                
		r8Uart115RxDelay <= 0;                 // RX delay                                     
		r4Bits115RxCount <= 0;                 // RX bits counter                              
		r8Uart115RxByte <= 0;                  // RX byte                                          
		bRx115ShiftReq <= 0;                   // Bit shift request                            
		rUdata115Out <= 0;                     // Received byte      
		rx_byte_cntr <= 0;							// Received byte counter
		zx_Keyboard_state <= 40'hffffffffff;	// 40 bit zx keyboard state, see keys.ods
		r8Key1 <= 0;									// First key
		r8Key2 <= 0;									// Second key
		r8Key3 <= 0;									// Third key
	end else
	begin
		case (r8Uart115RxState)                // Main state machine                           
			0:                                  // Default state, waiting for data              
			begin
				if (bkeyb_in == 0)               // if RX is 0 then it means that data is coming 
				begin
					r8Uart115RxNextState <= 1;    // Set next state var to next state             
					r8Uart115RxState <= 255;      // Jump to delay state next                     
					bRx115ShiftReq <= 0;          // There is no need to shift now                
					r8Uart115RxDelay <= 254;      // Set delay counter to 254(115200 baud)        
					r4Bits115RxCount <= 0;        // Reset bit counter                            
					r8Uart115RxByte <= 0;         // Clear RX byte                                
				end 
			end
			1:                                  // Receiving in progress                        
			begin
				if (r4Bits115RxCount < 7)        // If we still have some room to received       
				begin			
															// Increment bits counter                       
					r4Bits115RxCount <= r4Bits115RxCount + 1;
					bRx115ShiftReq <= 1;          // Shift will be required                       
					r8Uart115RxDelay <= 216;      // Set delay to 216                             
					r8Uart115RxState <= 255;      // Set state machine to delay state             
					r8Uart115RxNextState <= 1;    // Set state to jump to 1                       
					r8Uart115RxByte[7]<= bkeyb_in;// Get the bit                                  
				end else
				begin                            // Otherwise this is last bit                   
                                             // rUdataOut - is a received byte               
					rUdata115Out[7] <= bkeyb_in;  // Store it                                         
                                             // And store the rest of the bits               
					rUdata115Out[6:0] <= r8Uart115RxByte[6:0]; 
					r8Uart115RxState <= 2;        // Jump to next state                           
				end
			end
			2:                                  // State to set signal that byte is came        
			begin
				bRx115ShiftReq <= 0;             // Clear shift req                                   
															// If we are within the packet from keyboard
				if (rUdata115Out != 8'h57 && rUdata115Out != 8'hab && rUdata115Out != 8'h82 && rUdata115Out != 8'ha3)	
				begin
					if (rx_byte_cntr == 3)			// Shifts on the keyboard are presented as 3rd byte (count starts form 0)
					begin
															// Display shifts as byte 0
															// Caps shift
						if (rUdata115Out[3:0] == 4'h02)
						begin
															// 0 - if key is pressed
							zx_Keyboard_state[35] <= 0;
						end else
						begin
															// 1 - if key is released											
							zx_Keyboard_state[35] <= 1;
						end 
															// Symbol shift
						if (rUdata115Out[7:4] == 4'h02)
						begin
															// 0 - if key is pressed										
							zx_Keyboard_state[31] <= 0;
						end else
						begin
															// 1 - if key is released											
							zx_Keyboard_state[31] <= 1;
						end 
					end
					
					if (rx_byte_cntr == 5) 			// First pressed key comes as #5 in the sequence of bytes received
					begin	
															// Copy it into the appropriate memory loaction
						r8Key1 <= rUdata115Out[7:0];
					end
					
					if (rx_byte_cntr == 6) 			// Second pressed key comes as #6 in the sequence of bytes received
					begin
						r8Key2 <= rUdata115Out[7:0];
															// Copy it into the appropriate memory loaction
					end

					if (rx_byte_cntr == 7) 			// Third pressed key comes as #7 in the sequence of bytes received
					begin
															// Copy it into the appropriate memory loaction
						r8Key3 <= rUdata115Out[7:0];
					end
															// Actual keys presented as 5th/6th/7th byte (count starts from 0). So 8th state is when sequence complete
					if (rx_byte_cntr == 8)
					begin									
															// ZXCV
															// Z key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1d || r8Key2 == 8'h1d || r8Key3 == 8'h1d) 
						begin
							zx_Keyboard_state[36] <= 0;
						end else
						begin
							zx_Keyboard_state[36] <= 1;
						end 
															// X key, 0 pressed and 1 is released	
						if (r8Key1 == 8'h1b || r8Key2 == 8'h1b || r8Key3 == 8'h1b) //X
						begin
							zx_Keyboard_state[37] <= 0;
						end else
						begin
							zx_Keyboard_state[37] <= 1;
						end 
															// C key, 0 pressed and 1 is released	
						if (r8Key1 == 8'h06 || r8Key2 == 8'h06 || r8Key3 == 8'h06) 
						begin
							zx_Keyboard_state[38] <= 0;
						end else
						begin
							zx_Keyboard_state[38] <= 1;
						end 	
															// V key, 0 pressed and 1 is released
						if (r8Key1 == 8'h19 || r8Key2 == 8'h19 || r8Key3 == 8'h19) //V
						begin
							zx_Keyboard_state[39] <= 0;
						end else
						begin
							zx_Keyboard_state[39] <= 1;
						end 
															//	Space, symb, MNB
															// Space key, 0 pressed and 1 is released
						if (r8Key1 == 8'h2c || r8Key2 == 8'h2c || r8Key3 == 8'h2c) 
						begin
							zx_Keyboard_state[30] <= 0;
						end else
						begin
							zx_Keyboard_state[30] <= 1;
						end 	
															// M key, 0 pressed and 1 is released
						if (r8Key1 == 8'h10 || r8Key2 == 8'h10 || r8Key3 == 8'h10) 
						begin
							zx_Keyboard_state[32] <= 0;
						end else
						begin
							zx_Keyboard_state[32] <= 1;
						end 	
															// N key, 0 pressed and 1 is released
						if (r8Key1 == 8'h11 || r8Key2 == 8'h11 || r8Key3 == 8'h11) 
						begin
							zx_Keyboard_state[33] <= 0;
						end else
						begin
							zx_Keyboard_state[33] <= 1;
						end 	
															// B key, 0 pressed and 1 is released
						if (r8Key1 == 8'h05 || r8Key2 == 8'h05 || r8Key3 == 8'h05) 
						begin
							zx_Keyboard_state[34] <= 0;
						end else
						begin
							zx_Keyboard_state[34] <= 1;
						end 	
															// ASDFG
															// A key, 0 pressed and 1 is released
						if (r8Key1 == 8'h04 || r8Key2 == 8'h04 || r8Key3 == 8'h04) 
						begin
							zx_Keyboard_state[25] <= 0;
						end else
						begin
							zx_Keyboard_state[25] <= 1;
						end 	
															// S key, 0 pressed and 1 is released
						if (r8Key1 == 8'h16 || r8Key2 == 8'h16 || r8Key3 == 8'h16) 
						begin
							zx_Keyboard_state[26] <= 0;
						end else
						begin
							zx_Keyboard_state[26] <= 1;
						end 	
															// D key, 0 pressed and 1 is released
						if (r8Key1 == 8'h07 || r8Key2 == 8'h07 || r8Key3 == 8'h07) 
						begin
							zx_Keyboard_state[27] <= 0;
						end else
						begin
							zx_Keyboard_state[27] <= 1;
						end 	
															// F key, 0 pressed and 1 is released
						if (r8Key1 == 8'h09 || r8Key2 == 8'h09 || r8Key3 == 8'h09) 
						begin
							zx_Keyboard_state[28] <= 0;
						end else
						begin
							zx_Keyboard_state[28] <= 1;
						end 	
															// G key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0A || r8Key2 == 8'h0A || r8Key3 == 8'h0A) 
						begin
							zx_Keyboard_state[29] <= 0;
						end else
						begin
							zx_Keyboard_state[29] <= 1;
						end 	

															// HJKL ENT
															// H key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0B || r8Key2 == 8'h0B || r8Key3 == 8'h0B) 
						begin
							zx_Keyboard_state[24] <= 0;
						end else
						begin
							zx_Keyboard_state[24] <= 1;
						end 	
															// J key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0D || r8Key2 == 8'h0D || r8Key3 == 8'h0D) 
						begin
							zx_Keyboard_state[23] <= 0;
						end else
						begin
							zx_Keyboard_state[23] <= 1;
						end 	
															// K key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0E || r8Key2 == 8'h0E || r8Key3 == 8'h0E) 
						begin
							zx_Keyboard_state[22] <= 0;
						end else
						begin
							zx_Keyboard_state[22] <= 1;
						end 	
															// L key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0F || r8Key2 == 8'h0F || r8Key3 == 8'h0F) 
						begin
							zx_Keyboard_state[21] <= 0;
						end else
						begin
							zx_Keyboard_state[21] <= 1;
						end 	
															// Enter key, 0 pressed and 1 is released
						if (r8Key1 == 8'h28 || r8Key2 == 8'h28 || r8Key3 == 8'h28) 
						begin
							zx_Keyboard_state[20] <= 0;
						end else
						begin
							zx_Keyboard_state[20] <= 1;
						end 	

															// QWERT
															// Q key, 0 pressed and 1 is released
						if (r8Key1 == 8'h14 || r8Key2 == 8'h14 || r8Key3 == 8'h14) 
						begin
							zx_Keyboard_state[15] <= 0;
						end else
						begin
							zx_Keyboard_state[15] <= 1;
						end 	
															// W key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1a || r8Key2 == 8'h1a || r8Key3 == 8'h1a) 
						begin
							zx_Keyboard_state[16] <= 0;
						end else
						begin
							zx_Keyboard_state[16] <= 1;
						end 	
															// E key, 0 pressed and 1 is released
						if (r8Key1 == 8'h08 || r8Key2 == 8'h08 || r8Key3 == 8'h08) //E
						begin
							zx_Keyboard_state[17] <= 0;
						end else
						begin
							zx_Keyboard_state[17] <= 1;
						end 	
															// R key, 0 pressed and 1 is released
						if (r8Key1 == 8'h15 || r8Key2 == 8'h15 || r8Key3 == 8'h15) 
						begin
							zx_Keyboard_state[18] <= 0;
						end else
						begin
							zx_Keyboard_state[18] <= 1;
						end 	
															// T key, 0 pressed and 1 is released
						if (r8Key1 == 8'h17 || r8Key2 == 8'h17 || r8Key3 == 8'h17) 
						begin
							zx_Keyboard_state[19] <= 0;
						end else
						begin
							zx_Keyboard_state[19] <= 1;
						end 	

															// YUIOP
															// Y key, 0 pressed and 1 is released	
						if (r8Key1 == 8'h1c || r8Key2 == 8'h1c || r8Key3 == 8'h1c) 
						begin
							zx_Keyboard_state[14] <= 0;
						end else
						begin
							zx_Keyboard_state[14] <= 1;
						end 	
															// U key, 0 pressed and 1 is released
						if (r8Key1 == 8'h18 || r8Key2 == 8'h18 || r8Key3 == 8'h18) 
						begin
							zx_Keyboard_state[13] <= 0;
						end else
						begin
							zx_Keyboard_state[13] <= 1;
						end 	
															// I key, 0 pressed and 1 is released
						if (r8Key1 == 8'h0c || r8Key2 == 8'h0c || r8Key3 == 8'h0c) 
						begin
							zx_Keyboard_state[12] <= 0;
						end else
						begin
							zx_Keyboard_state[12] <= 1;
						end 	
															// O key, 0 pressed and 1 is released
						if (r8Key1 == 8'h12 || r8Key2 == 8'h12 || r8Key3 == 8'h12) 
						begin
							zx_Keyboard_state[11] <= 0;
						end else
						begin
							zx_Keyboard_state[11] <= 1;
						end 	
															// P key, 0 pressed and 1 is released
						if (r8Key1 == 8'h13 || r8Key2 == 8'h13 || r8Key3 == 8'h13) 
						begin
							zx_Keyboard_state[10] <= 0;
						end else
						begin
							zx_Keyboard_state[10] <= 1;
						end 	
															// 12345
															// 5 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h22 || r8Key2 == 8'h22 || r8Key3 == 8'h22) 
						begin
							zx_Keyboard_state[9] <= 0;
						end else
						begin
							zx_Keyboard_state[9] <= 1;
						end 	
															// 4 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h21 || r8Key2 == 8'h21 || r8Key3 == 8'h21) 
						begin
							zx_Keyboard_state[8] <= 0;
						end else
						begin
							zx_Keyboard_state[8] <= 1;
						end 	
															// 3 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h20 || r8Key2 == 8'h20 || r8Key3 == 8'h20) 
						begin
							zx_Keyboard_state[7] <= 0;
						end else
						begin
							zx_Keyboard_state[7] <= 1;
						end 	
															// 2 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1f || r8Key2 == 8'h1f || r8Key3 == 8'h1f) 
						begin
							zx_Keyboard_state[6] <= 0;
						end else
						begin
							zx_Keyboard_state[6] <= 1;
						end 	
															// 1 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h1e || r8Key2 == 8'h1e || r8Key3 == 8'h1e) 
						begin
							zx_Keyboard_state[5] <= 0;
						end else
						begin
							zx_Keyboard_state[5] <= 1;
						end 	
															// 67890
															// 6 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h23 || r8Key2 == 8'h23 || r8Key3 == 8'h23) 
						begin
							zx_Keyboard_state[4] <= 0;
						end else
						begin
							zx_Keyboard_state[4] <= 1;
						end 	
															// 7 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h24 || r8Key2 == 8'h24 || r8Key3 == 8'h24) 
						begin
							zx_Keyboard_state[3] <= 0;
						end else
						begin
							zx_Keyboard_state[3] <= 1;
						end 	
															// 8 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h25 || r8Key2 == 8'h25 || r8Key3 == 8'h25) 
						begin
							zx_Keyboard_state[2] <= 0;
						end else
						begin
							zx_Keyboard_state[2] <= 1;
						end 	
															// 9 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h26 || r8Key2 == 8'h26 || r8Key3 == 8'h26) 
						begin
							zx_Keyboard_state[1] <= 0;
						end else
						begin
							zx_Keyboard_state[1] <= 1;
						end 	
															// 0 key, 0 pressed and 1 is released
						if (r8Key1 == 8'h27 || r8Key2 == 8'h27 || r8Key3 == 8'h27) 
						begin
							zx_Keyboard_state[0] <= 0;
						end else
						begin
							zx_Keyboard_state[0] <= 1;
						end 	
					end
															// Increment bytes counter
					rx_byte_cntr = rx_byte_cntr + 1;
				end
				else begin								// If we are outside the packet then reset bytes counter
					rx_byte_cntr = 0;
				end
				r8Uart115RxDelay <= 216;         // One more delay will be required              
				r8Uart115RxState <= 255;         // So jump to delay state                       
				r8Uart115RxNextState <= 0;       // And set next counter to 0                    
			end 
			255:                                // Delay state                                  
			begin
				if (r8Uart115RxDelay != 0)       // If delay counter neq 0                       
				begin   	
															// Then decrement it
					r8Uart115RxDelay <= r8Uart115RxDelay - 1;                            
					if (bRx115ShiftReq)           // If shift was requested                       
					begin
						bRx115ShiftReq <= 0;       // Set shift req flag to 0                      
                                             // Actual shift                                 
						r8Uart115RxByte <= r8Uart115RxByte >> 1;   
					end
				end else                         // Otherwise                                    
				begin
															// Jump to the next required state              
					r8Uart115RxState <= r8Uart115RxNextState;
				end
			end
			default:                            // Preacution                                   
				begin
					r8Uart115RxState <= 0;        // State to 0                                   
				end
		endcase
	end
end

															// Main process for video
always@(posedge VideoClock or negedge videoreset)	
begin
	if (videoreset == 1'b0)								// If reset is active
	begin
		outVideoPCLK <= 0; 							// Set PCLK out to 0
		outPixCntr <= 0;								// Set Pixels in line counter to 0
		outVsync <= 0;									// Set output vsync to 0
		outHsync <= 0;									// Set output hsync to 0
		outLineCntr <= 0;								// Set lines in frame counter to 0
		out_B <= 0;										// Set B default to 0
		out_G <= 0;										// Set G default to 0
		out_R <= 0;										// Set R default to 0
		
		aVRAM_data_a0 <= 0;							// Reset lane a data
		aVRAM_wren_a0 <= 0;							// Reset write enable
		aVRAM_addr_a0 <= 0;							// Reset address	
		aVRAM_data_a1 <= 0;							// Reset lane a data
		aVRAM_wren_a1 <= 0;							// Reset write enable
		aVRAM_addr_a1 <= 0;							// Reset address	
		aVRAM_data_a2 <= 0;							// Reset lane a data
		aVRAM_wren_a2 <= 0;							// Reset write enable
		aVRAM_addr_a2 <= 0;							// Reset address	
		aVRAM_data_a3 <= 0;							// Reset lane a data
		aVRAM_wren_a3 <= 0;							// Reset write enable
		aVRAM_addr_a3 <= 0;							// Reset address	
		aVRAM_data_a4 <= 0;							// Reset lane a data
		aVRAM_wren_a4 <= 0;							// Reset write enable
		aVRAM_addr_a4 <= 0;							// Reset address	
		aVRAM_data_a5 <= 0;							// Reset lane a data
		aVRAM_wren_a5 <= 0;							// Reset write enable
		aVRAM_addr_a5 <= 0;							// Reset address	
		aVRAM_data_a6 <= 0;							// Reset lane a data
		aVRAM_wren_a6 <= 0;							// Reset write enable
		aVRAM_addr_a6 <= 0;							// Reset address	
		aVRAM_data_a7 <= 0;							// Reset lane a data
		aVRAM_wren_a7 <= 0;							// Reset write enable
		aVRAM_addr_a7 <= 0;							// Reset address	

		pixeladdr <= 0;								// Set plain pixel address (not in ZX way) to 0
		ColorLineAddr <= 0;							// Set color attribute counter (1 line only), range 0..31, used in calculating VRAM memory address
		ColorSync <= 0;								// Set flag to indicate even/odd pixel cylcle to give enough timeout to fetch data from color VRAM
		ColorDataMemCpy <= 0;						// Set color line (256 bits) to 0
		CurColorAttribute <= 0;						// Set current color attribute to 0
		CurColorLineAddr <= 0;						// Set address for attrubute cache to 0
		bDrawingBorder <= 0;							// Initialize drawing border flag
	end else												// if NOT reset
	begin												
		if (outVideoPCLK == 1)						// We need to devide 50M osc by 2 to get to 25M and assign active edge, so it will be H->L
		begin
			outVideoPCLK <= 0;						// Clear output PCLK
			case (outLineCntr)						// Check if it is time for Vsync
				`VSYNCSTART: begin					// If we are at the beggining of the frame then set Vsync to 0 as polarity is NEG
					outVsync <= 0;
				end
				`VSYNCEND: begin						// If we are at the end of the frame sync pulse then set Vsync to 1
					outVsync <= 1;
				end
			endcase
															// If we are in the active frame
			if (outPixCntr >= `HACTSTART && outPixCntr < `HACTEND && outLineCntr >= `VACTSTART && outLineCntr < `VACTEND)
			begin
															// ZX Spectrum screen is 256x192px, double that and it gives us 512x324 px
															// Hbackporch = (640-512)/2 = 64; Vbackporch = (480-324)/2 = 78
															// If we are at the edge the draw while rectangle  35+78  till 515 - 78  ie 384
				if (outPixCntr >= `HACTSTART + `HBACKPORCH && outPixCntr < `HACTEND - `HBACKPORCH && outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER)
				begin
					
					case (outPixCntr[3:0])			// If lsb pixel[2:0]*2 clocks == 0 then it means that we are at the beggining of new color space (8 pixels) on the screen
						0: begin
															// Fetching color attribute
							CurColorAttribute = ColorDataMemCpy[255:248];
						end
						1: begin
															// Fetching color attribute
							ColorDataMemCpy = ColorDataMemCpy << 8;
						end
					endcase
					
															// INK color based on video page pointer r8_128KSwitch[3] 
					if ((r8_128KSwitch[3] == 0 &&(aVRAM_q_a5[7 - outPixCntr[3:1]] == !(rZXFlashAttr & CurColorAttribute[7]))) || (r8_128KSwitch[3] == 1 &&(aVRAM_q_a7[7 - outPixCntr[3:1]] == !(rZXFlashAttr & CurColorAttribute[7]))))
					begin
															// Separate R, G and B, so this one is BLUE
						if ((CurColorAttribute & 1) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_B <= 4'b1111;		// Load bright blue
							end else
							begin
								out_B <= 4'b1011;   	// Load darker blue
							end
						end else
						begin
							out_B <= 0;			      // If there is no blue, then load blue = 0
						end
															// Separate R, G and B, so this one is RED
						if ((CurColorAttribute & 2) > 0)
						begin								
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_R <= 4'b1111;		// Load bright red
							end else
							begin
								out_R <= 4'b1011;		// Load darker red
							end
						end else
						begin
							out_R <= 0;       		// If there is no red, then load red = 0
						end
															// Separate R, G and B, so this one is GREEN
						if ((CurColorAttribute & 4) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin 
								out_G <= 4'b1111;		// Load bright green		
							end else
							begin
								out_G <= 4'b1011;		// Load darker green		
							end
						end else
						begin
							out_G <= 0;					// If there is no green, then load green = 0;
						end
						
					end else
					begin
															// PAPER color
															// Separate R, G and B, so this one is BLUE
						if ((CurColorAttribute & 8) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_B <= 4'b1111;   	// Load bright blue
							end else
							begin
								out_B <= 4'b1011;   	// Load darker blue
							end
						end else
						begin
							out_B <= 0;			      // If there is no blue, then 0
						end
															// Separate R, G and B, so this one is RED
						if ((CurColorAttribute & 16) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin
								out_R <= 4'b1111;		// Load bright red
							end else
							begin
								out_R <= 4'b1011;		// Load darker red		
							end
						end else
						begin
							out_R <= 0;					// If there is no red, then load 0
						end
															// Separate R, G and B, so this one is GREEN
						if ((CurColorAttribute & 32) > 0)
						begin
															// If bit 6 is set then it means that color is bright
							if ((CurColorAttribute & 64) > 0) 
							begin 
								out_G <= 4'b1111;  	// Set bright green
							end else
							begin
								out_G <= 4'b1011;		// Set darker green		
							end
						end else
						begin
							out_G <= 0;			      // If there is no gree, then set it to 0
						end
					end
					bDrawingBorder <= 0;				// We are not drawing border at the moment
				end else
				begin
															// Border color
					bDrawingBorder <= 1;	
					if (rBorderColor[0] == 1)		// Bit 0 is blue
					begin
						out_B <= 4'b1011;				// If it is set, then program out B to non-full brightness			
					end else
					begin
						out_B <= 4'b0000;				// If it is clear then program out B as no color			
					end
					
					if (rBorderColor[1] == 1) 		// Bit 1 is red
					begin
						out_R <= 4'b1011;				// If it is set, then program out R to non-full brightness			
					end else
					begin
						out_R <= 0;						// If it is clear then program out R as no color			
					end
					
					if (rBorderColor[2] == 1)		// Bit 2 is green
					begin
						out_G <= 4'b1011;				// If it is set, then program out G to non-full brightness			
					end else
					begin
						out_G <= 0;						// If it is clear then program out G as no color			
					end
				end
			end else
			begin
				out_B <= 0;								// If we are in the balnking area then assign B = 0
				out_G <= 0;								// If we are in the balnking area then assign G = 0
				out_R <= 0;								// If we are in the balnking area then assign R = 0
															// If we are within the active area vertically but in the backporch horizontally
															// then we need to pre-fetch color data from FPGA vram
				if (outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER && outPixCntr < `HACTSTART + `HBACKPORCH)
				begin
					if (ColorSync == 1)				// We need to do every other cycle to sync with FPGA RAM
					begin
															// If so, then copy data from FPGA RAM output to local 256 bit variable
															// pending a ZX128 video buffer adress check
						if (CurColorLineAddr < 32)
						begin
							if (r8_128KSwitch[3] == 0)
							begin
															// Load data from page 5
								ColorDataMemCpy = {ColorDataMemCpy[247:0], aVRAM_q_a5}; 
							end else
							begin
															// Load data from page 7
								ColorDataMemCpy = {ColorDataMemCpy[247:0], aVRAM_q_a7}; 
							end
															// And increment the pointer
							CurColorLineAddr = CurColorLineAddr + 1;
						end
					end 
				end 
			end											
															// Pre-fetching color data from FPGA RAM in vertical active area and a H backporch
			if (outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER && outPixCntr < `HACTSTART + `HBACKPORCH - 1)
			begin
				if (ColorSync == 0)					// If flag to sync (2x down clock speed) is eq 0
				begin
															// Calculate current address for color attribute
					if (r8_128KSwitch[3] == 0)    // r8_128KSwitch[3] tells us which video memory is used, 0 or 1, they are located in page 5 and 7
					begin
															// Set address for page 5
						aVRAM_addr_a5 = 6144 + ColorLineAddr + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 4)*32);
					end else
					begin
															// Set address for page 7
						aVRAM_addr_a7 = 6144 + ColorLineAddr + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 4)*32);
					end
					if (ColorLineAddr < 32)       // If there is next attrbute then
					begin
															// Increment current address
						ColorLineAddr = ColorLineAddr + 1;
					end
				end
			end else
			begin
				ColorLineAddr = 0;					// If we are out of color fetching area
				CurColorLineAddr = 0;				// Then set both counters, FPGA RAM address and 256 bit one to 0
			end
															//If NEXT pixel is in active area then setup address to read it
			if (outPixCntr >= `HACTSTART + `HBACKPORCH - 1 && outPixCntr < `HACTEND - `HBACKPORCH - 1 && outLineCntr >= `VACTSTART + `VBACKPORCH && outLineCntr < `VACTPAPER)
			begin
				pixeladdr = (((outPixCntr + 1 - `HACTSTART - `HBACKPORCH)>>1) >> 3) + (((outLineCntr - `VACTSTART - `VBACKPORCH) >> 1) * 32); 
															// Blocking assignments literally block the execution of the next
															// statement until the current statement is executed. (blocking '=', non blocking '<=')
															// Consequently, blocking assignments result in ordered
															// statement execution.
															// For example:
															// assume a = b = 0 initially;
															// a = 1; //executed first
															// b = a; //executed second
															// then a = 1, b = 1 after ordered execution. So:
															// ZX video memory is organized as following:
															// Y7  Y6  Y2  Y1  Y0 Y5  Y4 Y3 X4 X3 X2 X1 X0 ..or..
															// Y12 Y11 Y7  Y6  Y5 Y10 Y9 Y8 X4 X3 X2 X1 X0
				if (r8_128KSwitch[3] == 0)			// r8_128KSwitch[3] tells us which video memory is used, 0 or 1, they are located in page 5 and 7
				begin
															// Set address to page 5
					aVRAM_addr_a5 = pixeladdr[4:0] | pixeladdr[10:8] << 5 | pixeladdr[7:5] << 8 | pixeladdr [12:11] << 11;
				end else
				begin
															// Set address to page 7
					aVRAM_addr_a7 = pixeladdr[4:0] | pixeladdr[10:8] << 5 | pixeladdr[7:5] << 8 | pixeladdr [12:11] << 11;
				end 
			end

			case (outPixCntr)							// Check if it is time for Hsync
				`HSYNCSTART: begin					// If we are at the beginning of the line then set Hsync to 0 as polarity is NEG
					outHsync <= 0;
				end
				`HSYNCEND: begin						// If we are at the end of horizontal sync pulse then set Hsync to 1
					outHsync <= 1;
				end
			endcase
			
		end else											// If we are at the L->H PCLK satate
		begin
			ColorSync = ~ColorSync;
			outVideoPCLK = 1;							// Assign 1 to PCLK
			if (outPixCntr >= `HTOTAL - 1)		// If we are done with the line	
			begin
				outPixCntr <= 0;						// Set H pixel count to 0
				if (vsync_select)						// 50Hz mode selected
				begin
															// If we are done with the frame
					if (outLineCntr >= `VTOTAL50 - 1)
					begin
						outLineCntr <= 0;				// Set V counter to 0
					end else
					begin									// If not done then increment it
						outLineCntr <= outLineCntr + 1;
					end
				end else
				begin										// 60Hz mode selected
															// If we are done with the frame
					if (outLineCntr >= `VTOTAL60 - 1)
					begin
						outLineCntr <= 0;				// Set V counter to 0
					end else
					begin									// If not done then increment it
						outLineCntr <= outLineCntr + 1;
					end
				end
			end else
			begin
				outPixCntr <= outPixCntr + 1;		// If we are not done with the line then increment it
			end
		end
	end
end



always@(posedge outVsync or negedge rst_n)	// Flash counter, runs on Vsync
begin
	if (rst_n == 1'b0)								// If reset is active
	begin
		rZXFlashAttr <= 0;							// Set flash attrbute to 0
		rZXFlashCntr <= 0;							// Set flash counter to 0	
	end else
	begin
		if (rZXFlashCntr > `FLASHSPEED)			// If we reached 19 (ZX flash speed)
		begin
			rZXFlashCntr <= 0;						// Clear flash counter
			rZXFlashAttr <= ~rZXFlashAttr;		// Invert flash attribute	
		end else
		begin
			rZXFlashCntr <= rZXFlashCntr + 1;   // If we did not reach end, then increment the counter
		end
	end
end

always@(negedge rCpuClkMain or negedge rst_n)		// Interrupts on ZX Spectrum must happen everey 1/50S and last for 32 CPU clocks. 
begin
	if (rst_n == 1'b0)
	begin
		a_nINT <= 1;									// Init interrupt flag
		sp50Hzcntr <= 0;								// Reset the counter
		PreviousVsync <= 0;
	end else
	begin
		if (vsync_select)								// If true 50Hz is selected
		begin
															// If we are at the beginning of the frame
			if (outVsync == 0 && PreviousVsync == 1)
			begin
				sp50Hzcntr <= 0;						// Clear counter
				PreviousVsync <= 0;					// Clear previous vsync flag
			end else
			begin
				if (outVsync == 1)					// Wait till frame is active 
				begin		
					PreviousVsync <= 1;				// Set Vsync flag so we will know in the lines above if Vsync happen				
				end
															// If current pixel number is greater than 850 (VIDEO_PIX_INT) 
				if (outLineCntr * `HTOTAL + outPixCntr > `VIDEO_PIX_INT)
				begin
					case (sp50Hzcntr)					// Interrup shall take 31 CPU clocks
						0:  begin
							a_nINT <= 0;				// Interrupt flag set			
															// Increment counter
							sp50Hzcntr <= sp50Hzcntr + 1;		
						end
						31: begin						// Time to clear the interrupt
							a_nINT <= 1;				// Clear interrupt flag
															// Increment counter
							sp50Hzcntr <= sp50Hzcntr + 1;		
						end
						default: begin
															// Increment counter
							sp50Hzcntr <= sp50Hzcntr + 1;		
						end
					endcase
				end
			end
		end else
		begin												// If output is 60Hz then mimic 50Hz behaviour
			case (sp50Hzcntr)
				0:  begin
					a_nINT <= 0;						// Interrupt
					sp50Hzcntr <= sp50Hzcntr + 1;	// Increment counter
				end
				31: begin
					a_nINT <= 1;						// Clear interrupt flag
					sp50Hzcntr <= sp50Hzcntr + 1;	// Increment counter
				end
				70000: begin							// If 70000 is reached (50Hz = 70K CPU cycles)
					sp50Hzcntr <= 0;					// Then reset the counter
				end
				default: begin
					sp50Hzcntr <= sp50Hzcntr + 1;	// Increment counter
				end
			endcase
		end
	end
end

always@(negedge mc_clk or negedge rst_g)   	// Memory controller
begin
	if (rst_g == 1'b0)                        // In case of reset
	begin
		sound <= 1;										// Set sound enable bit
 		a_nWAIT <= 1;									// CPU signals, no wait			
 		a_nNMI <= 1;									//	CPU signals, no unmaskable interrupt		
 		a_nBUSRQ <= 1;									// CPU signals, no BUSREQ
		aROM_addr <= 0;								// ROM address is 0 at start
		r8_128KSwitch <= 0;							// Memory switch byte is equal 0 at strat
		DataToDisplay <= 0;							// Debug
		bPrevTin <= tapein;							// Tape in, initialize to current state
		bTinLock <= 0;									// Switch to turn on tape in is off
		rCounterUnlockTape <= 0;					// Counter to unlock tape in	
	end else
	begin
		if (tapein != bPrevTin)						// Mimic enflow of EAR signal to tape in 
		begin
			bTinLock <= 1;								// Tape in lock if input change
			bPrevTin <= tapein;						// Record it for future
			rCounterUnlockTape <= 0;				// Reset counter
		end else
		begin
			bPrevTin <= tapein;						// Record for future use
															// If timer has expired 
			if (rCounterUnlockTape == 22'b1111111111111111111111)
			begin
				bTinLock <= 0;							// Remove tape lock
			end else										// otherwise
			begin
															// Increment counter
				rCounterUnlockTape <= rCounterUnlockTape + 1;
			end
		end
		if (a_nIORQ == 0)								// If we are in the port access mode			
		begin
			if (a_nWR == 0)							// If we are writing then on ZX spectrum the only port to write is 0xFE, there are no other ports
			begin
															// If we are writing to a port 0xFE (see ZX spectrum documentation) 
				if (wZXGlobalAddress[7:0] == 8'hfe)															
				begin
					rBorderColor[2:0] <= a_D[2:0];// Set the border color, bits 2:0
					sound <= a_D[4];					// Set the sound bit, bit 4
				end else
				begin										// If write is to address 7ffd then it is 128K memory control  byte
					if (wZXGlobalAddress[15] == 0 && wZXGlobalAddress[1] == 0 && r8_128KSwitch[5] == 0)
					begin
						r8_128KSwitch <= a_D;		// Update memory control from CPU bus
						DataToDisplay <= a_D;		// Debug
					end else
					begin
						if (wZXGlobalAddress[15] == 1 && wZXGlobalAddress[1] == 0)
						begin
							AY8910_d_in <= a_D;		// AY8910 write
						end
					end
				end
			end else
			begin
				if (a_nRD == 0)						// If we are reading port 
				begin
															// If it is AY8910
					if (wZXGlobalAddress[15] == 1 && wZXGlobalAddress[1] == 0)
					begin
															// Then provide AY8910 data
						wZXGlobalDataO[7:0] <= AY8910_d_out;
					end else
					begin
						case(wZXGlobalAddress[7:0])// If it is no AY8910 then lets find out what it is
						8'hfe:							// Port FE	
						begin
															// Mimic issue of tape in availability for ear output
							wZXGlobalDataO[5] <= sound; 
							if (bTinLock ==  1) 		// If it is enabled
							begin
															// Then we load tape data
								wZXGlobalDataO[6] <= ~tapein;
							end else
							begin
															// Otherwise load sound out
								wZXGlobalDataO[6] <= sound;
							end
															// And bit 7 always load the sound out
							wZXGlobalDataO[7] <= sound;		

							case (wZXGlobalAddress)	// We need to see which one we are reading
								16'hfefe: begin		// Port fefe - keyobard inputs, caps thru Z
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[39:35];
								end						// Port 7ffe
								16'h7ffe: begin		// Keys BNM, sym shift and space
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[34:30];
								end
								16'hfdfe: begin		// Keys ASDFG
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[29:25];
								end
								16'hbffe: begin		// Keys HJKL Enter
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[24:20];
								end
								16'hfbfe: begin		// Keys QWERT
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[19:15];
								end
								16'hdffe: begin		// Keys YUIOP
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[14:10];
								end
								16'hf7fe: begin		// Keys 12345
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[9:5];
								end
								16'heffe: begin		// Keys 67890
									wZXGlobalDataO[4:0] <= zx_Keyboard_state[4:0];
								end
								default: begin			// 8 bit address port readings		
									            		// For all other ports ??FE we need to do the following
															// "ANY" key simulation in case if program polls port 254 as indication of such event
															// If no key pressed
									if (zx_Keyboard_state == 40'hffffffffff)
									begin
															// ...then load all ones
										wZXGlobalDataO[4:0] <= 5'b11111;
									end else 
									begin
															// Otherwise simulate "any" key
										wZXGlobalDataO[4:0] <= 5'b11110;
									end
								end
							endcase //case (wZXGlobalAddress)
						end
						8'h1F: begin
							wZXGlobalDataO[7:5] <= 3'b000;
															// Bits are inverted
															// Right - bit 0
															// Left  - bit 1
															// Up    - bit 3
															// Down  - bit 2
															// Fire  - bit 4
							wZXGlobalDataO[4:0] <= ~KempstonJoystick[4:0];
						end
						default:							// Special feature to emulate ULA memory leakage
							begin
								if (bDrawingBorder == 0)
								begin						// In the active area non used ports shall return current byte being displayed
									if (r8_128KSwitch[3] == 0)
									begin					// Active page is 5
										wZXGlobalDataO[7:0] <= aVRAM_q_a5;
									end else
									begin					// Active page is 7
										wZXGlobalDataO[7:0] <= aVRAM_q_a7;
									end
								end else
								begin
															// While it should return FF if drawing the border
									wZXGlobalDataO[7:0] <= 8'hFF;
								end
							end
						endcase //end of case(wZXGlobalAddress[7:0]))
					end//end of if (wZXGlobalAddress[15] == 1 && wZXGlobalAddress[1] == 0)
				end //end of a_nRD = 0
			end //end ofif (a_nWR == 0)
		end else //end of if (a_nIORQ == 0)
		begin
			if (wZXGlobalAddress< `ZX_RAM_START)// If address is less than ZX_RAM_START, then it is ROM access
			begin
															// Reading ROM, address set
				aROM_addr[13:0] <= wZXGlobalAddress[13:0];	
															// MSB of address is a bit 4 of r8_128KSwitch
				aROM_addr[14] <= r8_128KSwitch[4];
				wZXGlobalDataO <= aROM_q;			// Reading ROM, setting output data bus to a ROM data bus value
			end else
			begin
															// Main memory paging
				if (wZXGlobalAddress < `RAM_PAGE1)
				begin	
															// Page from 16384 to 32768 maps to ram 5
					aVRAM_addr_b5 <= wZXGlobalAddress - 16384;
					aVRAM_wren_b5 <= ~a_nWR;		// Set WR signal	
					wZXGlobalDataO <= aVRAM_q_b5;	// Set data out	
					aVRAM_data_b5 <= a_D;  			// Set data in 	
				end else
				begin
					if (wZXGlobalAddress < `RAM_PAGE2)
					begin
															// Page from 32768 to 49152 maps to ram 2
						aVRAM_addr_b2 <= wZXGlobalAddress - `RAM_PAGE1;
						aVRAM_wren_b2 <= ~a_nWR;	// Set WR signal
															// Set data out
						wZXGlobalDataO <= aVRAM_q_b2;
						aVRAM_data_b2 <= a_D;  		// Set data in		
					end else
					begin
						case (r8_128KSwitch[2:0])  // 128K specific pages, address is 49152-65536
							0: begin
								aVRAM_addr_b0 <= wZXGlobalAddress - `RAM_PAGE2;
															// Page 0 WR
								aVRAM_wren_b0 <= ~a_nWR;
															// Page 0 data out
								wZXGlobalDataO <= aVRAM_q_b0;		
								aVRAM_data_b0 <= a_D;// Page 0 data in  				
							end
							1: begin						// Same as above
								aVRAM_addr_b1 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b1 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b1;		
								aVRAM_data_b1 <= a_D;// Same as above  				
							end
							2: begin						// Same as above
								aVRAM_addr_b2 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b2 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b2;		
								aVRAM_data_b2 <= a_D;// Same as above  				
							end
							3: begin						// Same as above
								aVRAM_addr_b3 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b3 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b3;		
								aVRAM_data_b3 <= a_D;// Same as above  				
							end
							4: begin						// Same as above
								aVRAM_addr_b4 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b4 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b4;		
								aVRAM_data_b4 <= a_D;// Same as above  				
							end
							5: begin						// Same as above
								aVRAM_addr_b5 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b5 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b5;		
								aVRAM_data_b5 <= a_D;// Same as above  				
							end 
							6: begin						// Same as above
								aVRAM_addr_b6 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b6 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b6;		
								aVRAM_data_b6 <= a_D;// Same as above  				
							end
							7: begin						// Same as above
								aVRAM_addr_b7 <= wZXGlobalAddress - `RAM_PAGE2;
															// Same as above
								aVRAM_wren_b7 <= ~a_nWR;
															// Same as above
								wZXGlobalDataO <= aVRAM_q_b7;		
								aVRAM_data_b7 <= a_D;// Same as above  				
							end
						endcase
					end
				end	
			end
		end
	end
end


always @(negedge clk50)								// Blinking LED, for fun
begin
	if (BlLedDivider == 0) 							// Blinking LED, for fun
	begin
		BlLedDivider <= 1;							// Blinking LED, for fun
		led0 <= ~led0;									// Blinking LED, for fun
	end else												// Blinking LED, for fun
	begin
		BlLedDivider <= BlLedDivider + 1;		// Blinking LED, for fun
	end 
end

always @(negedge clk50 or negedge rst_n)		//  Reducing a speed for led to 25MHz
begin
	if (rst_n == 1'b0)								// In case of reset
	begin
		clk25 <= 0;										// Assign clk 25 to 0
	end else
	begin
		clk25 =~clk25;									// Invert clk25
	end
end

always@(negedge clk25 or negedge rst_n)      // 7+1 digit control
begin
	if (rst_n == 0)									// In case of reset
	begin
		CurDigit <= 0;									// Set CurDigit to 0
	end else
	begin
		case (NextDigit)
			8'h0: begin									// In case of 0 to be displayed
					CurDigit <= 8'b11000000;		// Load a bit value to be shown on the LED into register
				end
			8'h1: begin									// In case of 1 to be displayed
					CurDigit <= 8'b11111001;		// Load a bit value to be shown on the LED into register
				end
			8'h2: begin									// In case of 2 to be displayed
					CurDigit <= 8'b10100100;		// Load a bit value to be shown on the LED into register
				end
			8'h3: begin									// In case of 3 to be displayed
					CurDigit <= 8'b10110000;		// Load a bit value to be shown on the LED into register
				end
			8'h4: begin									// In case of 4 to be displayed
					CurDigit <= 8'b10011001;		// Load a bit value to be shown on the LED into register
				end
			8'h5: begin									// In case of 5 to be displayed
					CurDigit <= 8'b10010010;		// Load a bit value to be shown on the LED into register
				end
			8'h6: begin									// In case of 6 to be displayed
					CurDigit <= 8'b10000010;		// Load a bit value to be shown on the LED into register
				end
			8'h7: begin									// In case of 7 to be displayed
					CurDigit <= 8'b11111000;		// Load a bit value to be shown on the LED into register
				end
			8'h8: begin									// In case of 8 to be displayed
					CurDigit <= 8'b10000000;		// Load a bit value to be shown on the LED into register
				end
			8'h9: begin									// In case of 9 to be displayed
					CurDigit <= 8'b10010000;		// Load a bit value to be shown on the LED into register
				end
			8'ha: begin									// In case of a to be displayed
					CurDigit <= 8'b10001000;		// Load a bit value to be shown on the LED into register
				end
			8'hb: begin									// In case of b to be displayed
					CurDigit <= 8'b10000011;		// Load a bit value to be shown on the LED into register
				end
			8'hc: begin									// In case of c to be displayed
					CurDigit <= 8'b10100111;		// Load a bit value to be shown on the LED into register
				end
			8'hd: begin									// In case of d to be displayed
					CurDigit <= 8'b10100001;		// Load a bit value to be shown on the LED into register
				end
			8'he: begin									// In case of e to be displayed
					CurDigit <= 8'b10000110;		// Load a bit value to be shown on the LED into register
				end
			8'hf: begin									// In case of f to be displayed
					CurDigit <= 8'b10001110;		// Load a bit value to be shown on the LED into register
				end
		endcase
	end
end

always@(posedge clk25 or negedge rst_n)		// Dislay data on 6x(7+1) indicator
begin
	if (rst_n == 0)									// In case of reset
	begin
		tstDATA_LED <= 48'hFFFFFF;					// Output: 6 idicators by 8 LEDs, total 48 bits
		DigSelect <= 0;								// Select eq 0
		NextDigit <= 0;								// Next digit eq 0
	end else
	begin
		case (DigSelect)								// Based on which indicator we are updating now
			0: begin										// Digit #0
															// Bits 7:0, load current digit
					tstDATA_LED  [7:0] <= CurDigit;
					DigSelect <= 1;					// Next is digit #1
															// Load next digit to decode
					NextDigit[3:0] <= DataToDisplay[7:4]; 
				end
			1: begin										// Digit #1
					tstDATA_LED [15:8] <= CurDigit;
					DigSelect <= 2;					// Next is digit #2	
															// Load next digit to decode
					NextDigit[3:0] <= DataToDisplay[11:8];
				end
			2: begin										// Digit #1
					tstDATA_LED[23:16] <= CurDigit;
					DigSelect <= 3;					// Next is digit #3
															// Load next digit to decode
					NextDigit[3:0] <= DataToDisplay[15:12]; 
				end
			3: begin										// Digit #3
					tstDATA_LED[31:24] <= CurDigit;
					DigSelect <= 4;					// Next is digit #4
															// Load next digit to decode
					NextDigit[3:0] <= DataToDisplay[19:16]; 
				end
			4: begin										// Digit #4
					tstDATA_LED[39:32] <= CurDigit;
					DigSelect <= 5;					// Next is digit #5
															// Load next digit to decode
					NextDigit[3:0] <= DataToDisplay[23:20];
				end
			5: begin										// Digit #5
					tstDATA_LED[47:40] <= CurDigit;
					DigSelect <= 0;					// Next is digit #0
															// Load next digit to decode
					NextDigit[3:0] <= DataToDisplay[3:0];
				end
		endcase
	end
end

endmodule